/*
 * STATsrv: STAT-service daemon 
 *          part of dim_STAT tool <http://dimitrik.free.fr>
 *
 * (c) 1997-2007 by (dim) Dimitri KRAVTCHUK <dimitri.kravtchuk <at> free.fr>
 *
 ***************************************************************************
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published  by  the *
 * Free Software Foundation; either version 2 of the License, or (at  your *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it  will  be  useful,  but *
 * WITHOUT ANY WARRANTY; without the implied warranty  of  MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License *
 * for more details.                                                       *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                   *
 ***************************************************************************
 */



#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>           /* struct timeval */

struct sockaddr_in sinme;
struct sockaddr_in sinhim;
struct sockaddr_in frominet;

extern int errno;
void err();
void mes();
int Nread();
int Nwrite();

#define VERSION      "v.3.4 (dim)"
#define ACCESS_NAME  "/etc/STATsrv/access"

void sigpipe() {  signal(SIGPIPE, sigpipe); }

void sigchld()
{
 int sst;

 wait( &sst );
 signal( SIGCHLD, sigchld );
}

int fd= -1;                         /* fd of network socket */
int fd_cmd= -1;

void sigterm()
{
 mes( "*** EXIT: got SIGTERM! ***" );

 if( fd_cmd != -1 )  close( fd_cmd );
 if( fd != -1 )  close( fd );
 kill( 0, SIGTERM );

 exit(0);
}

Usage( name ) 
char *name;
{ fprintf( stderr, "%s %s\nUsage: %s [-p port (5000 default)]\n", name, VERSION, name ); exit( -2 ); }



main( argc, argv )
int argc;
char *argv[];
{
 int domain, fromlen;
 short port= 5000;               /* TCP port number */
 int i, cnt;

 struct sockaddr_in peer;
 int peerlen= sizeof(peer);

 if( access( ACCESS_NAME, 0 ) != 0 )  { printf( "Not found access file: %s\n", ACCESS_NAME ); exit( -2 ); }

 for( i= 1; i < argc; i++ )
 {
  if( strcmp( argv[i], "-p" ) == 0 )
  {
   port= atoi( argv[++i] );
   continue;
  }

  Usage( argv[0] );
 }

 setpgrp();


 sinme.sin_port=  htons(port);
#if defined(__convex__)
 sinme.sin_family= AF_INET;
 sinme.sin_addr.s_addr= INADDR_ANY ;
#endif

 while(1)
 {
  if((fd= socket(AF_INET, SOCK_STREAM, 0)) < 0 )    err("socket");
  if( bind( fd, &sinme, sizeof(sinme) ) >= 0 )  break;
  mes( "-- Bind Error, will try again..." );
  sleep( 30 );
 }

 signal(SIGPIPE, sigpipe);
 signal(SIGCHLD, sigchld);
 signal(SIGTERM, sigterm);

 listen(fd,0);   /* allow a queue of 0 */
 mes( "*** READY for Service! ***" );

 fromlen= sizeof(frominet);
 domain= AF_INET;

 while( 1 )
 {
  if((fd_cmd= accept( fd, &frominet, &fromlen )) < 0 )
  {
   if( errno == EINTR )  continue;
   err("accept");
  }

  if( fork() == 0 )
  {
   if( getpeername( fd_cmd, (struct sockaddr_in *) &peer, &peerlen ) < 0 )  err("getpeername");
   STATsrv( fd_cmd,  inet_ntoa(peer.sin_addr) );
   close( fd_cmd );

   exit(0);
  }

  close( fd_cmd );
 }
}


FILE *check_command( cmdline, addr, fd )
char *cmdline;
char *addr;
int fd;
{
 FILE *fp;
 char cmd[100], msg[200], buff[512], *chp, *tail;
 char key[20], key_cmd[40], key_fullcmd[100];
 int i, n, access, Ok, found, List= 0;

 /* 
  * remove ; and | symbols to avoid hidden commands 
  * probably some other?...
  */
 while( (chp= (char *)strchr( cmdline, ';' )) != NULL )  *chp= ' ';
 while( (chp= (char *)strchr( cmdline, '|' )) != NULL )  *chp= ' ';

 for( cmd[0]= '\0', i= n= 0; i < sizeof( cmd ) && cmdline[n] != '\0'; n++ )
 {
  if( cmdline[n] == ' ' )
  {
   if( i > 0 )  break;
   continue;
  }

  cmd[i++]= cmdline[n];
 }

 cmd[i]= '\0';
 tail= &cmdline[n];
 while( (chp= (char *)strchr( cmd, '/' )) != NULL )  strcpy( cmd, chp+1 );

 if( strcmp( cmd, "STAT_LIST" ) == 0 )  
 {
  sprintf( msg, "STAT *** LIST COMMAND (%s)\n", cmd );
  Nwrite( fd, msg, strlen( msg ) );
  List= 1;
 }
 
 if( (fp= fopen( ACCESS_NAME, "r" )) == NULL )
 {
  sprintf( msg, "STAT *** BAD COMMAND (STAT service not configured: %s)\n", cmd );
  Nwrite( fd, msg, strlen( msg ) );
  return( NULL );
 }

 for( access= found= 0, Ok= 1; fgets( buff, sizeof(buff), fp ) != NULL; )
 {
  if( buff[0] == '#' )  continue;
  if( strlen( buff ) < 5 )  continue;
  sscanf( buff, "%s%s%s", key, key_cmd, key_fullcmd );

  if( strcmp( key, "access" ) == 0 )
  {
   if( access && Ok )  continue;
   access= 1;
   Ok= 0;
   if( strcmp( addr, key_cmd ) == 0 )  Ok= 1;
   continue;
  }

  if( strcmp( key, "command" ) == 0 )
  {
   access= 0;
   if( !Ok )  continue;
   if( List )
   {  
    sprintf( msg, "STAT: %s\n", key_cmd );
    Nwrite( fd, msg, strlen( msg ) );
    continue;
   }

   if( strcmp( cmd, key_cmd ) == 0 )  { found= 1; break; }
  }
 }

 fclose( fp );

 if( List )
 {
  sprintf( msg, "STAT *** LIST END (%s)\n", cmd );
  Nwrite( fd, msg, strlen( msg ) );
  return( NULL );
 }

 if( !found )
 {
  sprintf( msg, "STAT *** BAD COMMAND (no access: %s)\n", cmd );
  Nwrite( fd, msg, strlen( msg ) );
  return( NULL );
 }

 /* avoid buff overflow */
 strcpy( buff, key_fullcmd );
 strcat( buff, " " );
 strncat( buff, tail, sizeof(buff) - strlen(buff) - 1 );
 buff[ sizeof(buff)-1 ]= '\0';
 
 if( (fp= popen( buff, "r" )) == NULL )
 {
  sprintf( msg, "STAT *** BAD COMMAND (cannot start: %s)\n", cmd );
  Nwrite( fd, msg, strlen( msg ) );
  return( fp );
 }

 if( fgets( buff, sizeof(buff), fp ) == NULL )
 {
  sprintf( msg, "STAT *** BAD COMMAND (not found: %s)\n", cmd );
  Nwrite( fd, msg, strlen( msg ) );
  pclose( fp );
  return( NULL );
 }

 sprintf( msg, "STAT *** OK COMMAND (cmd: %s)\n", cmd );
 Nwrite( fd, msg, strlen( msg ) );
 Nwrite( fd, buff, strlen( buff ) );
 return( fp );
}


#define P_SIZE 256
#define B_SIZE 400

kill_chld()
{
 FILE *pp;
 char buff[ B_SIZE + 8];
 long pid;
 long plist[ P_SIZE ];
 int  n;
 int  Ok;

 pid= getpid();
 Ok= 0;
 
 while( !Ok )
 {
  sprintf( buff, "/etc/STATsrv/bin/ptree %d", pid );
  pp= popen( buff, "r" );
 
  while( fgets( buff, B_SIZE, pp ) != NULL )
  {
   sscanf( buff, "%d", &plist[0] );
   if( plist[0] == pid )  break;
  }
  
  for( n= 0; n < P_SIZE && fgets( buff, B_SIZE, pp ) != NULL; n++ )  
   sscanf( buff, "%d", &plist[n] );
   
  if( n < P_SIZE )  Ok= 1;
  while( n > 0 )  kill( plist[--n], SIGKILL );
 }
}


STATsrv( fd, addr )
int fd;
char *addr;
{
 char cmd[48], buff[1024], *msg;
 long T;
 int cnt;
 FILE *pp;

 errno= 0;
 if( (cnt= Nread( fd, buff, sizeof(buff) )) <= 0 )  err( "Nread" );
 time( &T );
 fprintf( stderr, "STATsrv-%d:> Connect -> [%s] CMD= \"%s\" TIME: %s", getpid(), addr, buff, ctime(&T) );
 fflush( stderr );

 if( (pp= check_command( buff, addr, fd )) != NULL )
 {
  strncpy( cmd, buff, 40 );
  buff[40]= '\0';
  if( strlen( buff ) > 40 )  strcat( cmd, " ..." );

  while( fgets( buff, sizeof( buff ), pp ) != NULL )
   if( Nwrite( fd, buff, strlen( buff ) ) <= 0 )
   {
    // kill child procs!!
    kill_chld();
    pclose( pp );
    break;
   }
 }

 close( fd );
 time( &T );
 fprintf( stderr, "STATsrv-%d:> Exit -> [%s] CMD= \"%s\" TIME: %s", getpid(), addr, cmd, ctime(&T) );
 fflush( stderr );
}


void err( s )
char *s;
{
 perror(s);
 fprintf( stderr," errno= %d\n", errno );
 fflush( stderr );
 exit( -2 );
}

void mes( s )
char *s;
{
 fprintf( stderr,"STATsrv-%d: %s\n", getpid(), s );
 fflush( stderr );
}



Nread( fd, buf, count )
int fd;
char *buf;
int count;
{
 int cnt;
 cnt= read( fd, buf, count );
 if( cnt >= 0 ) buf[cnt]= 0;
 return( cnt );
}

Nwrite( fd, buf, count )
int fd;
char *buf;
int count;
{
 return( write( fd, buf, count ) );
}

